--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        has_flags.lua
--

-- imports
import("core.cache.detectcache")
import("core.language.language")

-- is linker?
function _islinker(flags, opt)
    -- the tool kind is gcld or gcsh?
    local toolkind = opt.toolkind or ""
    return toolkind:endswith("ld") or toolkind:endswith("sh")
end

-- try running
function _try_running(program, argv, opt)
    local errors = nil
    local ok = try { 
        function () 
            os.runv(program, argv, opt)
            return true 
        end, 
        catch { 
            function (errs) 
                errors = (errs or ""):trim() 
            end 
        }
    }
    return ok, errors
end

-- attempt to check it from the argument list
function _check_from_arglist(flags, opt, islinker)
    -- only for compiler and single flag
    if islinker or #flags > 1 then
        return
    end

    -- make cache key
    local key = "core.tools.go.has_flags"

    -- make flags key
    local flagskey = opt.program .. "_" .. (opt.programver or "")

    -- get all flags from argument list
    local allflags = detectcache:get2(key, flagskey)
    if not allflags then
        allflags = {}
        
        -- try to get help from "go help build"
        local help_output = try { 
            function () 
                return os.iorunv(opt.program, {"help", "build"}, {envs = opt.envs}) 
            end 
        }
        
        if help_output then
            -- parse flags from help output
            for flag in help_output:gmatch("%s+(%-[%-%a%d=]+)%s+") do
                allflags[flag] = true
            end
        end
        
        -- also try "go build -h" for more detailed flags
        local build_help = try { 
            function () 
                return os.iorunv(opt.program, {"build", "-h"}, {envs = opt.envs}) 
            end 
        }
        
        if build_help then
            for flag in build_help:gmatch("%s+(%-[%-%a%d=]+)%s+") do
                allflags[flag] = true
            end
        end

        -- save cache
        detectcache:set2(key, flagskey, allflags)
    end
    
    return allflags[flags[1]]
end

-- try running to check flags
function _check_try_running(flags, opt, islinker)
    -- make an stub source file
    local sourcefile = path.join(os.tmpdir(), "detect", "go_has_flags.go")
    if not os.isfile(sourcefile) then
        io.writefile(sourcefile, "package main\n\nfunc main() {\n}\n")
    end

    -- check flags for linker or compiler
    -- Modern Go uses "go build" for both compilation and linking
    local argv = {"build", "-o", os.tmpfile()}
    table.join2(argv, flags)
    table.insert(argv, sourcefile)
    
    local ok, errors = _try_running(opt.program, argv, {envs = opt.envs})
    return ok == true
end

-- has_flags(flags)?
--
-- @param opt   the argument options, e.g. {toolname = "", program = "", programver = "", toolkind = "[gc|gcld|gcar]"}
--
-- @return      true or false
--
function main(flags, opt)
    -- is linker?
    local islinker = _islinker(flags, opt)

    -- attempt to check it from the argument list
    if _check_from_arglist(flags, opt, islinker) then
        return true
    end

    -- try running to check it
    return _check_try_running(flags, opt, islinker)
end
