--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        xmake.lua
--

-- define rule: cuda.build
rule("cuda.build")
    set_sourcekinds("cu")
    add_deps("cuda.build.devlink")
    on_build_files("private.action.build.object", {jobgraph = true, batch = true})
    on_config(function (target)
        -- https://github.com/xmake-io/xmake/issues/4755
        local cu_ccbin = target:tool("cu-ccbin")
        if cu_ccbin then
            target:add("cuflags", "-ccbin=" .. os.args(cu_ccbin), {force = true})
            target:add("culdflags", "-ccbin=" .. os.args(cu_ccbin), {force = true})
        end
    end)

-- define rule: cuda
rule("cuda")

    -- add build rules
    add_deps("cuda.build", "cuda.gencodes")

    -- inherit links and linkdirs of all dependent targets by default
    add_deps("utils.inherit.links")

