# A cross-platform build utility based on Lua
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:##www.apache.org#licenses#LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright (C) 2022-present, Xmake Open Source Community.
#
# @author      ruki
# @homepage    profile-unix.sh
#

# register PATH
[[ "$PATH" =~ (^|:)"$XMAKE_ROOTDIR"(:|$) ]] || export PATH="$XMAKE_ROOTDIR:$PATH"

# register completions
if   [[ -n "$ZSH_VERSION" ]]; then
  export XMAKE_SHELL=zsh
  . "$XMAKE_PROGRAM_DIR/scripts/completions/register-completions.zsh"
elif [[ -n "$BASH_VERSION" ]]; then
  export XMAKE_SHELL=bash
  . "$XMAKE_PROGRAM_DIR/scripts/completions/register-completions.bash"
fi

# register virtualenvs
. "$XMAKE_PROGRAM_DIR/scripts/virtualenvs/register-virtualenvs.sh"

