//
//  AppDelegate.h
//  test5
//
//  Created by ruki on 2020/4/8.
//  Copyright © 2020 tboox. All rights reserved.
//

#import <UIKit/UIKit.h>

@interface AppDelegate : UIResponder <UIApplicationDelegate>


@end

