--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      ruki
-- @file        cosmoar.lua
--

inherit("ar")

function init(self)
    _super.init(self)
end

function link(self, objectfiles, targetkind, targetfile, flags, opt)
    opt = opt or {}
    if is_host("windows") then
        targetfile = targetfile:gsub("\\", "/")
        local objectfiles_new = {}
        for idx, objectfile in ipairs(objectfiles) do
            objectfiles_new[idx] = objectfiles[idx]:gsub("\\", "/")
        end
        objectfiles = objectfiles_new
    end
    return _super.link(self, objectfiles, targetkind, targetfile, flags, table.join(opt, {shell = true}))
end
